// Copyright 2024 The Perses Authors
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package common

import (
	"encoding/json"
	"net/url"
	"path"
)

func ParseURL(rawURL string) (*URL, error) {
	u, err := url.Parse(rawURL)
	if err != nil {
		return nil, err
	}
	return &URL{URL: u}, nil
}

func MustParseURL(rawURL string) *URL {
	u, err := url.Parse(rawURL)
	if err != nil {
		panic(err)
	}
	return &URL{URL: u}
}

func NewURL(u *URL, paths ...string) *URL {
	result := &URL{
		URL: &url.URL{},
	}
	if u == nil {
		return result
	}
	*result.URL = *u.URL
	result.Path = path.Join(paths...)
	return result
}

type URL struct {
	*url.URL `json:"-" yaml:"-"`
}

func (u *URL) IsNilOrEmpty() bool {
	return u.URL == nil || u.URL.String() == ""
}

// UnmarshalYAML implements the yaml.Unmarshaler interface for URLs.
func (u *URL) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var s string
	if err := unmarshal(&s); err != nil {
		return err
	}

	urlp, err := url.Parse(s)
	if err != nil {
		return err
	}
	u.URL = urlp
	return nil
}

// MarshalYAML implements the yaml.Marshaler interface for URLs.
func (u URL) MarshalYAML() (interface{}, error) {
	if u.URL != nil {
		return u.String(), nil
	}
	return nil, nil
}

// UnmarshalJSON implements the json.Marshaler interface for URL.
func (u *URL) UnmarshalJSON(data []byte) error {
	var s string
	if err := json.Unmarshal(data, &s); err != nil {
		return err
	}
	urlp, err := url.Parse(s)
	if err != nil {
		return err
	}
	u.URL = urlp
	return nil
}

// MarshalJSON implements the json.Marshaler interface for URL.
func (u URL) MarshalJSON() ([]byte, error) {
	if u.URL != nil {
		return json.Marshal(u.URL.String())
	}
	return []byte("null"), nil
}

// MarshalText implements the encoding.TextMarshaler interface.
func (u *URL) MarshalText() ([]byte, error) {
	return []byte(u.URL.String()), nil
}

// UnmarshalText implements the encoding.TextUnmarshaler interface.
func (u *URL) UnmarshalText(text []byte) error {
	urlp, err := url.Parse(string(text))
	if err != nil {
		return err
	}
	u.URL = urlp
	return nil
}
